import { Vector2, Vector3 } from 'three'
import { Line, Polyline } from 'cadapi'
import type { PlaceBlock } from '../vo/order/PlaceBlock'
import type { SaleBlockHole } from '../vo/model/SaleBlockHole'
import type { SaleBlockModel } from '../vo/model/SaleBlockModel'
import { Point } from '../vo/point/Point'
import { BlockPlus } from '../vo/order/BlockPlus'
import { Arc2d, CADExt } from './base/CAD'

/** 图形类，两种图形一种异形和一种矩形 */
class Graphy
{
  private Polyline: Polyline

  private _IsRect: boolean = undefined
  /** 判断是否是盒子图形，水平竖直矩形 */
  get IsRect()
  {
    if (this._IsRect == undefined)
    {
      for (let index = 0; index < this.AllPoints.length - 1; index++)
      {
        const pt1 = this.AllPoints[index]
        const pt2 = this.AllPoints[index + 1]
        if (Math.abs(pt1.X - pt2.X) > 0.1 && Math.abs(pt1.Y - pt2.Y) > 0.1)
        {
          this._IsRect = false
          return
        }
      }
      for (let pt of this.AllPoints)
      {
        if (!this.Box.onLinePt(pt.X, pt.Y))
        {
          this._IsRect = false
          return
        }
      }
      this._IsRect = true
    }

    return this._IsRect
  }

  private _AllPoints: CfPoint[]
  /** 所有的夹点 */
  get AllPoints()
  {
    if (this._AllPoints == undefined)
    {
      this._AllPoints = []
      for (let i = 0; i < this.Polyline.LineData.length - 1; i++)
      {
        let prePt = this.Polyline.LineData[i]
        let lastPt = this.Polyline.LineData[i + 1]
        if (prePt.bul != 0)
        {
          let points = CADExt.getPointsFromCurve(prePt.pt.x, prePt.pt.y, lastPt.pt.x, lastPt.pt.y, prePt.bul, 2)
          for (let pt of points)
          {
            this._AllPoints.push(new CfPoint(pt.x, pt.y))
          }
        } else
        {
          this._AllPoints.push(new CfPoint(prePt.pt.x, prePt.pt.y))
        }
      }
    }
    return this._AllPoints
  }

  private _Box: Box
  /** 外包框 盒子对象 */
  get Box()
  {
    if (this._Box == undefined)
    {
      let boud = this.Polyline.BoundingBox
      this._Box = Box.createBox(new CfPoint(boud.min.x, boud.min.y), new CfPoint(boud.max.x, boud.max.y))
    }
    return this._Box
  }

  private constructor(pl: Polyline)
  {
    this.Polyline = pl
  }

  /** 通过盒子创建图形 */
  public static createByBox(pBox: Box): Graphy
  {
    let pl: Polyline = new Polyline()
    let points = pBox.get4Pt()
    points.push(points[0])

    for (let i = 0; i < points.length; i++)
    {
      const pt = points[i]
      pl.AddVertexAt(i, new Vector2(pt.X, pt.Y))
    }

    return Graphy.createByPolyline(pl)
  }

  public static createByPolyline(pl: Polyline): Graphy
  {
    let g = new Graphy(pl)
    return g
  }

  /** 判断两个图形是否相交 */
  public intersectGraphy(inGraphy: Graphy): boolean
  {
    if (this.Box.intersect(inGraphy.Box))
    {
      let v3d = this.Polyline.IntersectWith(inGraphy.Polyline, 0)
      if (v3d.length > 0)
        return true
    }
    return false
  }

  /** 判断两个图形是否相交 */
  public intersectBox(inBox: Box): boolean
  {
    if (this.Box.intersect(inBox))
    {
      // if (this.IsRect) return true;
      let gra = Graphy.createByBox(inBox)
      return this.intersectGraphy(gra)
    }
    return false
  }

  /** 判断两个图形是否覆盖 */
  public coverBox(inBox: Box): boolean
  {
    if (this.Box.intersect(inBox))
    {
      if (this.IsRect)
        return true
      // 判断矩形点是否在图形内部
      let pts = inBox.get4Pt()
      for (let pt of pts)
      {
        if (this.includesPt(pt))
          return true
      }
      // 判断异形点是否在矩形内部
      for (let pt of this.AllPoints)
      {
        if (inBox.includesPt(pt.X, pt.Y))
          return true
      }
      // 两个图形相交则覆盖
      let gra = Graphy.createByBox(inBox)
      return this.intersectGraphy(gra)
    }

    return false
  }

  /** 判断点是否在轮廓内部 */
  public includesPt(pt: CfPoint): boolean
  {
    if (this.Box.includesPt(pt.X, pt.Y))
    {
      if (this.IsRect)
        return true// 矩形
      // 异形
      let v: Vector3 = new Vector3(pt.X, pt.Y)
      return this.Polyline.PtInCurve(v)
    }
    return false
  }

  /** 判断box是否在轮廓内部 */
  public includesBox(box: Box): boolean
  {
    if (!this.includesPt(box.minPt))
      return false
    if (!this.includesPt(box.maxPt))
      return false

    return true
  }

  /** 判断box是否在轮廓内部 */
  public includesGraphy(box: Graphy): boolean
  {
    for (let pt of box.AllPoints)
    {
      if (!this.includesPt(pt))
        return false
    }

    return true
  }

  /** 获取最近的延伸相交点 */
  public getCloseExtendIntersectPoint(pt: CfPoint): CfLine[]
  {
    let reLes: CfLine[] = []
    let v3d = new Vector3(pt.X, pt.Y, 0)
    let closePt = this.Polyline.GetClosestPointTo(v3d, false)
    let tempClosePt = new CfPoint(closePt.x, closePt.y)

    let vectors = this.intersectLine(new CfPoint(v3d.x, v3d.y), tempClosePt)
    if (vectors.length < 2)
      return reLes
    for (let i = 0; i <= vectors.length - 2; i++)
    {
      // 数组里面的第一个点为最近点
      let pt1 = vectors[i]
      let pt2 = vectors[i + 1]
      if (tempClosePt.DistensTo(pt1) < tempClosePt.DistensTo(pt2))
      {
        let le: CfLine = new CfLine(pt1, pt2)
        reLes.push(le)
      } else
      {
        let le: CfLine = new CfLine(pt2, pt1)
        reLes.push(le)
      }
      i += 1
    }

    return reLes
  }

  /** 图形相交于线返回多个点 */
  public intersectLine(sPt: CfPoint, ePt: CfPoint): CfPoint[]
  {
    // 此处只获取相交内部的线段
    let line = new Line(new Vector3(sPt.X, sPt.Y), new Vector3(ePt.X, ePt.Y))
    let vectors = this.Polyline.IntersectWith(line, 2)
    // 相交出来的可能是个切点所以有可能会有奇数出现
    let rePoint: CfPoint[] = []
    for (let i = 0; i <= vectors.length - 2; i++)
    {
      let pt1 = vectors[i]
      let pt2 = vectors[i + 1]
      let cpt = new CfPoint((pt1.x + pt2.x) / 2, (pt1.y + pt2.y) / 2)
      if (this.includesPt(cpt))
      {
        rePoint.push(new CfPoint(pt1.x, pt1.y))
        rePoint.push(new CfPoint(pt2.x, pt2.y))
        i += 1
      }
    }

    return rePoint
  }
}

/** 标签盒子类 */
export class NoteBox
{
  private _Pt: CfPoint
  private _W: number
  private _L: number
  /** 中点 */
  get CenterPt() { return this._Pt }
  /** 宽 */
  get W() { return this._W }
  /** 长 */
  get L() { return this._L }

  private _box: Box
  /** 外包框 矩形 */
  get Box()
  {
    if (this._box == undefined)
    {
      let minPt = new CfPoint(this._Pt.X - this.W / 2, this._Pt.Y - this.L / 2)
      let maxPt = new CfPoint(this._Pt.X + this.W / 2, this._Pt.Y + this.L / 2)
      this._box = Box.createBox(minPt, maxPt)
    }
    return this._box
  }

  constructor(pt: CfPoint, w: number, l: number)
  {
    this._Pt = pt
    this._W = w
    this._L = l
  }

  /** 判断是否包含点位 */
  public getIncludesPoints(points: CfPoint[]): CfPoint[]
  {
    let reArrPt: CfPoint[] = []
    for (let pt of points)
    {
      if (this.Box.includesPt(pt.X, pt.Y))
        reArrPt.push(pt)
    }
    return reArrPt
  }

  /** 判断是否包含点位 */
  public includesPoints(points: CfPoint[]): boolean
  {
    for (let pt of points)
    {
      if (this.Box.includesPt(pt.X, pt.Y))
        return true
    }
    return false
  }

  /** 移动盒子对象 */
  public moveBox(pt: CfPoint)
  {
    if (pt.X < 0)
    {
      let T = 0
    }
    this._Pt = pt
    this._box = undefined
  }
}

class CfLine
{
  private _Pt1: CfPoint
  private _Pt2: CfPoint

  get StartPt()
  {
    return this._Pt1
  }

  get EndPt()
  {
    return this._Pt2
  }

  public constructor(pt1: CfPoint, pt2: CfPoint)
  {
    if (pt1 == undefined)
    {
      let jj = 0
    }
    this._Pt1 = pt1
    this._Pt2 = pt2
  }

  get Length() { return this._Pt1.DistensTo(this._Pt2) }

  getCenterPt()
  {
    return this.StartPt.extendPoint(this.EndPt, this.Length / 2)
  }

  /** 排序 */
  public innerOrderPoints(points: CfPoint[]): CfPoint[]
  {
    let rePoint: CfPoint[] = []
    rePoint.push(this.StartPt)
    while (points.length > 0)
    {
      let index = 0
      let minLen: number = Number.MAX_VALUE
      for (let i = 0; i < points.length; i++)
      {
        const element = points[i]
        let len = this.StartPt.DistensTo(element)
        if (minLen > len)
        {
          index = i
          minLen = len
          break
        }
      }
      rePoint.push(points[index])
      points.splice(index, 1)
    }

    rePoint.push(this.EndPt)
    return rePoint
  }

  /** 旋转线 */
  public rotate(centerPt: CfPoint, ang: number): CfLine
  {
    let pt1 = this.StartPt.Rotate(centerPt, ang)
    let pt2 = this.EndPt.Rotate(centerPt, ang)
    return new CfLine(pt1, pt2)
  }
}

/** 点 */
class CfPoint
{
  public X = 0
  public Y = 0

  get length()
  {
    return Math.sqrt(this.X * this.X + this.Y * this.Y)
  }

  constructor(x: number, y: number)
  {
    this.X = x
    this.Y = y
  }

  /** 延伸一个点 */
  public extendPoint(pt: CfPoint, length: number): CfPoint
  {
    let vert1 = this.Subtract(pt).Normal()
    let tempPt = vert1.Multiply(length)
    return this.Add(tempPt)
  }

  /** 延伸一个点 */
  public extendVector(vector: CfPoint, length: number): CfPoint
  {
    let vert1 = vector.Normal()
    let tempPt = vert1.Multiply(length)
    return this.Add(tempPt)
  }

  public Add(p1: CfPoint): CfPoint
  {
    return new CfPoint(p1.X + this.X, p1.Y + this.Y)
  }

  /** otherPt - 本身 */
  public Subtract(otherPt: CfPoint): CfPoint
  {
    return new CfPoint(otherPt.X - this.X, otherPt.Y - this.Y)
  }

  public DistensTo(pt: CfPoint): number
  {
    let tempX = pt.X - this.X
    let tempy = pt.Y - this.Y
    let len = Math.sqrt(tempX * tempX + tempy * tempy)
    return len
  }

  public Normal(): CfPoint
  {
    let k = 1 / this.length
    return new CfPoint(this.X * k, this.Y * k)
  }

  public Multiply(c: number): CfPoint
  {
    return new CfPoint(this.X * c, this.Y * c)
  }

  public Rotate(centerPt: CfPoint, ang: number): CfPoint
  {
    let thisPt = new Vector2(this.X, this.Y)
    let vert1 = new Vector2(centerPt.X, centerPt.Y)
    let tempPt = thisPt.rotateAround(vert1, ang)
    return new CfPoint(tempPt.x, tempPt.y)
  }
}

/** 图形碰撞类 */
export class GraphAvoidance
{
  private Block: PlaceBlock
  // 标签的宽
  private NoteW: number
  // 标签的长
  private NoteL: number
  // 小板外轮廓图形对象
  private BlockGraph: Graphy
  // 造型数组对象
  private ModelesGraph: Graphy[] = []

  public static debugId: string = ''

  constructor(b: PlaceBlock, w: number, l: number)
  {
    this.Block = b
    this.NoteW = w
    this.NoteL = l
    let points = this.getBorderPoints(b)
    let pl: Polyline = new Polyline()
    for (let i = 0; i < points.length; i++)
    {
      const pt = points[i]
      pl.AddVertexAt(i, new Vector2(pt.X, pt.Y))
    }
    this.BlockGraph = Graphy.createByPolyline(pl)
  }

  /** 获取自动标签的计算点位 */
  public getAutoPrintNotePoint(isFaceB: boolean = false, avoidHole: boolean = true): NoteBox
  {
    let centerPt = this.BlockGraph.Box.getCenterPt()
    let nodeBox: NoteBox = new NoteBox(centerPt, this.NoteW, this.NoteL)
    if (nodeBox.L > this.BlockGraph.Box.L - 2 && nodeBox.W > this.BlockGraph.Box.W - 2)
      return nodeBox

    // 获取图形数据

    let boxs: Graphy[] = []
    if (avoidHole)
    {
      // 初始化模型和造型数据
      boxs = this.initModelAndHole(isFaceB)
    }

    // 判断是否可以循环打印
    if (this.canWhilePrintNote(nodeBox, boxs))
    {
      // 是否对齐居中点位
      this.isAlignCenterNote(nodeBox, boxs)
      return nodeBox
    } else
    {
      // 获取无法找到打印点的节点
      return this.getNoPrintNotePoint(boxs)
    }
  }

  /** 大多数情况为板很小的时候调用 */
  private getNoPrintNotePoint(boxs: Graphy[]): NoteBox
  {
    if (this.BlockGraph.IsRect && this.ModelesGraph.length == 0 && boxs.length == 0)// 矩形无法找到是就贴中间
    {
      let centerPt = this.BlockGraph.Box.getCenterPt()
      return new NoteBox(centerPt, this.NoteW, this.NoteL)
    } else // 异形需要找点位
    {
      let tempPt = this.lookUpPoint(boxs)
      if (tempPt != undefined)
        return new NoteBox(tempPt, this.NoteW, this.NoteL)
    }
    return undefined
  }

  /** 寻找点位 */
  private lookUpPoint(boxs: Graphy[]): CfPoint
  {
    let centerPt = this.BlockGraph.Box.getCenterPt()
    let note = new NoteBox(centerPt, this.NoteW, this.NoteL)
    let val: number = 2
    let maxLen = Math.max(this.BlockGraph.Box.L, this.BlockGraph.Box.W) / 2
    // 偏移寻找
    while (true)
    {
      if (val >= maxLen)
        break
      let tempLes: CfLine[] = getArrayCurve(centerPt, val)
      val += 2
      let ArrLine: CfLine[] = []
      for (let le of tempLes)
      {
        let line = this.lookUpLine(le.StartPt, le.EndPt)
        if (line == undefined)
          continue
        note.moveBox(line.getCenterPt())
        if (this.isPrintNote(note, boxs, [], false))
        {
          ArrLine.push(line)
        }
      }
      if (ArrLine.length == 0)
        continue
      let tempLe = ArrLine.sort((a, b) => b.Length - a.Length)[0]
      return tempLe.getCenterPt()
    }

    // 对角寻找不管哪种只要相交的都算
    let points = this.BlockGraph.Box.get4Pt()
    let upPt = new CfPoint(centerPt.X, centerPt.Y + 100)
    let leftPt = new CfPoint(centerPt.X + 100, centerPt.Y)
    let ArrLine: CfLine[] = []
    let le = this.lookUpLine(centerPt, upPt)
    if (le != undefined)
      ArrLine.push(le)
    le = this.lookUpLine(centerPt, leftPt)
    if (le != undefined)
      ArrLine.push(le)
    le = this.lookUpLine(points[0], points[2])
    if (le != undefined)
      ArrLine.push(le)
    le = this.lookUpLine(points[1], points[3])
    if (le != undefined)
      ArrLine.push(le)

    // 获取长度最大的线
    let tempLe = ArrLine.sort((a, b) => b.Length - a.Length)[0]
    if (tempLe == undefined)
      return centerPt
    return tempLe.getCenterPt()

    function getArrayCurve(centerPt: CfPoint, val: number): CfLine[]
    {
      let tempLes: CfLine[] = []
      tempLes.push(new CfLine(new CfPoint(centerPt.X - val, centerPt.Y + 100), new CfPoint(centerPt.X - val, centerPt.Y)))
      tempLes.push(new CfLine(new CfPoint(centerPt.X + val, centerPt.Y + 100), new CfPoint(centerPt.X + val, centerPt.Y)))

      tempLes.push(new CfLine(new CfPoint(centerPt.X + 100, centerPt.Y + val), new CfPoint(centerPt.X, centerPt.Y + val)))
      tempLes.push(new CfLine(new CfPoint(centerPt.X + 100, centerPt.Y - val), new CfPoint(centerPt.X, centerPt.Y - val)))

      return tempLes
    }
  }

  /** 寻找最长的线 */
  private lookUpLine(pt1: CfPoint, pt2: CfPoint): CfLine
  {
    let profileGrapy = this.BlockGraph
    let arrPt = profileGrapy.intersectLine(pt1, pt2)
    if (arrPt.length < 2)
      return undefined
    let tempLine = new CfLine(arrPt[0], arrPt[1])

    let rePoints: CfPoint[] = []
    for (let g of this.ModelesGraph)
    {
      let temps = g.intersectLine(pt1, pt2)
      if (temps.length > 1 && temps.length % 2 == 0)
      {
        temps.forEach(t => rePoints.push(t))
      }
    }
    let points = tempLine.innerOrderPoints(rePoints)

    let ArrLine: CfLine[] = []
    for (let i = 0; i <= points.length - 2; i++)
    {
      AddLine(points[i], points[i + 1])
      i += 1
    }

    // 获取长度最大的线
    let tempLe = ArrLine.sort((a, b) => b.Length - a.Length)[0]
    return tempLe

    function AddLine(pt1: CfPoint, pt2: CfPoint)
    {
      let le1 = new CfLine(pt1, pt2)
      ArrLine.push(le1)
    }
  }

  /** 是否对齐居中点位 */
  private isAlignCenterNote(nodeBox: NoteBox, graphys: Graphy[])
  {
    if (this.BlockGraph.IsRect)
      return// 矩形不做判断
    let cPt = nodeBox.CenterPt
    // 获取最近点的延伸相交的两个点
    let les = this.BlockGraph.getCloseExtendIntersectPoint(cPt)
    if (les.length == 0)
      return
    for (let le of les)
    {
      if (le.StartPt.DistensTo(cPt) > this.NoteL * 2)
        continue
      if (le.Length < this.NoteL * 1.5)
        continue // 距离过小的不做判断
      let tempLe = le.rotate(le.getCenterPt(), Math.PI / 2)
      let points = this.BlockGraph.intersectLine(tempLe.StartPt, tempLe.EndPt)
      for (let i = 0; i <= points.length - 2; i++)
      {
        let le4 = new CfLine(points[i], points[i + 1])
        nodeBox.moveBox(le4.getCenterPt())
        if (this.isPrintNote(nodeBox, graphys, []))
          return
      }
    }

    for (let le of les)
    {
      if (le.StartPt.DistensTo(cPt) > this.NoteL * 2)
        continue
      if (le.Length < this.NoteL * 1.5)
        continue // 距离过小的不做判断
      nodeBox.moveBox(le.getCenterPt())
      if (this.isPrintNote(nodeBox, graphys, []))
        return
    }

    nodeBox.moveBox(cPt)// 将点位还原
  }

  /** 初始化模型和造型数据 */
  private initModelAndHole(isFaceB: boolean): Graphy[]
  {
    let boxs: Graphy[] = []
    let mGs = this.ModelesGraph

    if (this.Block.IsScrapBlock)
      return boxs
    let holes = this.getHoles(this.Block, isFaceB)// BlockHelper.GetHoles_DoFace(this.Block, isFaceB);
    for (let h of holes)// 孔数据
    {
      let pt: CfPoint = new CfPoint(h.PointX, h.PointY)
      let pBox = Box.createBoxByPoint(pt, h.Radius)
      boxs.push(Graphy.createByBox(pBox))
    }

    // 造型数据
    let models = this.getModels(this.Block, isFaceB)// BlockHelper.GetModels_DoFace(this.Block, isFaceB);
    let vms: SaleBlockModel[] = []
    for (let model of models)
    {
      if (model.isVKnifeModel) {
        vms.push(model)
        continue
      }
      if (model.HasContour)// 普通造型
      {
        let pts = model.OriginModeling.outline.pts
        let buls = model.OriginModeling.outline.buls
        let pl: Polyline = new Polyline()
        for (let i = 0; i < pts.length; i++)
        {
          let v = new Vector2(pts[i].x, pts[i].y)
          pl.AddVertexAt(i, v)
          pl.SetBulgeAt(i, buls[i])
        }
        Add(Graphy.createByPolyline(pl))
      } else
      {
        let pl: Polyline = new Polyline()
        for (let i = 0; i < model.PointList.length; i++)
        {
          let theP = model.PointList[i]
          let v = new Vector2(theP.PointX, theP.PointY)
          pl.AddVertexAt(i, v)
          pl.SetBulgeAt(i, theP.Curve)
        }
        Add(Graphy.createByPolyline(pl))
      }
    }
    // 二维刀路
    // let vms = BlockHelper.getVModels(this.Block);
    for (let vm of vms)
    {
      // if (models.includes(vm.model) == false) continue;
      for (let vl of vm.VLines) {
        let points = vl.points
        // 普通造型
        for (let i = 0; i < points.length - 1; i++)
        {
          let prevP = points[i]
          let theP = points[i + 1]
          let pl: Polyline = this.getOffsetPolyline(prevP, theP, vl.knifeRadius, vl.knifeRadius)
          let tempG = Graphy.createByPolyline(pl)
          Add(tempG)
        }
      }
    }

    return boxs

    function Add(g: Graphy)
    {
      if (g.Box.L < 50 && g.Box.W < 50)
      {
        boxs.push(g)
      } else
      {
        for (let i = 0; i < mGs.length; i++)
        {
          let tempG = mGs[i]
          if (tempG.includesGraphy(g))
            return
          if (g.includesGraphy(tempG))
          {
            mGs[i] = g
            return
          }
        }
        mGs.push(g)
      }
    }
  }

  /** 循环移动此节点判断是否可以打印 */
  private canWhilePrintNote(nodeBox: NoteBox, graphys: Graphy[]): boolean
  {
    let firstBox: Graphy[] = []
    // 第一次判断
    if (this.isPrintNote(nodeBox, graphys, firstBox))
      return true
    //  return true;
    // 外轮廓循环寻找
    let filterBox: Box = nodeBox.Box
    if (firstBox.length > 0)
    {
      filterBox = Box.createBoxByBoxs(firstBox)
    }
    // 循环
    while (true)
    {
      let reBoxs: Graphy[] = []
      if (this.canMovePrintNote(nodeBox, graphys, filterBox, reBoxs))
        return true
      if (reBoxs.length == 0)
      {
        if (filterBox.W >= this.Block.PlaceWidth && filterBox.L >= this.Block.PlaceLength)
          break
        // filterBox.extendBox(nodeBox.W, nodeBox.L);
        filterBox.extendBox(1, 1)
        continue
      }
      let tempBox = Box.createBoxByBoxs(reBoxs)
      filterBox.AddBox(tempBox)
    }

    // 通过造型边界判断
    if (this.ModelesGraph.length > 0)
    {
      for (let g of this.ModelesGraph)
      {
        let cpt: CfPoint = g.Box.getCenterPt()
        nodeBox.moveBox(cpt)
        filterBox = g.Box
        if (this.canMovePrintNote(nodeBox, graphys, filterBox, []))
          return true
      }
    }

    // 通过圆孔边界寻找
    if (graphys.length > 0)
    {
      for (let g of firstBox)
      {
        let cpt: CfPoint = g.Box.getCenterPt()
        nodeBox.moveBox(cpt)
        filterBox = g.Box
        if (this.canMovePrintNote(nodeBox, graphys, filterBox, []))
          return true
      }

      for (let g of graphys)
      {
        let cpt: CfPoint = g.Box.getCenterPt()
        nodeBox.moveBox(cpt)
        filterBox = g.Box
        if (this.canMovePrintNote(nodeBox, graphys, filterBox, []))
          return true
      }
    }
    //

    return false
  }

  /** 此节点移动完之后是否可以打印 */
  private canMovePrintNote(nodeBox: NoteBox, rects: Graphy[], filterBox: Box, reBoxs: Graphy[]): boolean
  {
    let tempPts = this.getFilterPoint(this.Block, nodeBox, filterBox)
    for (let pt of tempPts)
    {
      let cpt = filterBox.getCenterPt()
      let len = cpt.DistensTo(pt)
      let newPt = cpt.extendPoint(pt, len + nodeBox.W / 2)
      if (!GraphAvoidance.checkPoint(this.Block, newPt.X, newPt.Y))
        continue
      nodeBox.moveBox(newPt)
      if (this.isPrintNote(nodeBox, rects, reBoxs))
      {
        reBoxs.length = 0// 清空数组
        return true
      }
    }
    return false
  }

  /** 标签是否可以打印 */
  private isPrintNote(nodeBox: NoteBox, rects: Graphy[], reBox: Graphy[], breakProfile = true): boolean
  {
    if (this.isNodeInProfile(nodeBox, breakProfile))
    {
      let arrPt = this.getIntersectBox(nodeBox, rects)
      if (arrPt.length == 0)
        return true
      arrPt.forEach(t => reBox.push(t))
    }
    return false
  }

  /** 判断点是否在轮廓内部 */
  private isNodeInProfile(nodeBox: NoteBox, breakProfile = true): boolean
  {
    if (breakProfile)
    {
      if (!this.BlockGraph.includesBox(nodeBox.Box))
        return false// 不包含盒子
      if (this.BlockGraph.intersectBox(nodeBox.Box))
        return false// 盒子和图形相交
    }

    // 造型是否相交于造型
    if (this.ModelesGraph.length > 0)
    {
      for (let m of this.ModelesGraph)
      {
        if (m.coverBox(nodeBox.Box))
        {
          let jj = 0
          return false
        }
      }
    }
    return true
  }

  /** 获取过滤的点数据 */
  private getFilterPoint(b: PlaceBlock, nodeBox: NoteBox, box: Box): CfPoint[]
  {
    let cpt = box.getCenterPt()// 盒子的中点
    let arrPt: CfPoint[] = []
    let xlen = box.W / 2 + nodeBox.W / 2
    let ylen = box.L / 2 + nodeBox.L / 2
    pushPoint(cpt.X + xlen, cpt.Y)// 添加右侧点
    pushPoint(cpt.X - xlen, cpt.Y)// 添加左侧点
    if (box.L > nodeBox.L)
    {
      addPointx(nodeBox, box, xlen)
      addPointx(nodeBox, box, -xlen)
    }

    pushPoint(cpt.X, cpt.Y + ylen)// 添加上部点
    pushPoint(cpt.X, cpt.Y - ylen)// 添加下部点
    if (box.W > nodeBox.W)
    {
      addPointy(nodeBox, box, ylen)
      addPointy(nodeBox, box, -ylen)
    }
    if (box.W < nodeBox.W || box.L < nodeBox.L)
    {
      let arrPt = box.get4Pt()
      pushDiagonalPoint(arrPt[0], arrPt[2])// 斜右上
      pushDiagonalPoint(arrPt[2], arrPt[0])
      pushDiagonalPoint(arrPt[1], arrPt[3])
      pushDiagonalPoint(arrPt[3], arrPt[1])
    }
    return arrPt

    function addPointx(nodeBox: NoteBox, box: Box, xlen: number)
    {
      let cpt = box.getCenterPt()// 盒子的中点
      let count = Math.ceil(box.L / nodeBox.L)
      let numLen: number = nodeBox.L
      for (let i = 1; i < count; i++)
      {
        pushPoint(cpt.X + xlen, cpt.Y + numLen)// 添加左右侧点
        numLen = -numLen
        if (numLen > 0)
        {
          numLen = numLen + nodeBox.L
        }
      }
    }
    function addPointy(nodeBox: NoteBox, box: Box, ylen: number)
    {
      let count = Math.ceil(box.W / nodeBox.W)
      let numLen: number = nodeBox.W
      for (let i = 1; i < count; i++)
      {
        pushPoint(cpt.X + numLen, cpt.Y + ylen)// 添加上下点
        numLen = -numLen
        if (numLen > 0)
        {
          numLen = numLen + nodeBox.W
        }
      }
    }

    function pushDiagonalPoint(sPt: CfPoint, ePt: CfPoint)
    {
      let len = nodeBox.Box.getDiagonalLength() / 2 + box.getDiagonalLength()
      let pt = sPt.extendPoint(ePt, len)
      pushPoint(pt.X, pt.Y)
    }
    function pushPoint(x: number, y: number)
    {
      if (!GraphAvoidance.checkPoint(b, x, y))
        return
      arrPt.push(new CfPoint(x, y))// 添加上部点
    }
  }

  /** 获取造型数据 */
  private getModels(block: PlaceBlock, isFaceB: boolean): SaleBlockModel[]
  {
    let modelsT = block.Models_OrgTrhough.slice()
    if (isFaceB)
    {
      let modelsB = (block.IsTurnOver ? block.Models_OrgFaceA.slice() : block.Models_OrgFaceB.slice())
      modelsT = modelsT.concat(modelsB)
    } else
    {
      let modelsA = (block.IsTurnOver ? block.Models_OrgFaceB.slice() : block.Models_OrgFaceA.slice())
      modelsT = modelsT.concat(modelsA)
    }
    return modelsT
  }

  /** 获取孔数据 */
  private getHoles(block: PlaceBlock, isFaceB: boolean): SaleBlockHole[]
  {
    let HolesT = block.Holes_OrgThrough.slice()
    if (isFaceB)
    {
      let holesB = (block.IsTurnOver ? block.Holes_OrgFaceA : block.Holes_OrgFaceB)
      HolesT = HolesT.concat(holesB)
    } else
    {
      let holesA = (block.IsTurnOver ? block.Holes_OrgFaceB : block.Holes_OrgFaceA)
      HolesT = HolesT.concat(holesA)
    }
    return HolesT
  }

  /** 获取偏移轮廓 */
  private getOffsetPolyline(prevP, theP, leftDis: number, rightDis: number): Polyline
  {
    let pl: Polyline = new Polyline()
    let vpt1 = new Vector2(prevP.x, prevP.y)
    let vpt2 = new Vector2(theP.x, theP.y)
    pl.AddVertexAt(0, vpt1)
    pl.AddVertexAt(1, vpt2)
    pl.SetBulgeAt(0, prevP.bul)

    let temppls = pl.GetOffsetCurves(leftDis)[0]
    let temppls2 = pl.GetOffsetCurves(-rightDis)[0]

    temppls2.AddVertexAt(2, temppls.GetPoint2dAt(1))
    temppls2.AddVertexAt(3, temppls.GetPoint2dAt(0))
    temppls2.SetBulgeAt(2, -temppls.GetBulgeAt(0))
    temppls2.AddVertexAt(4, temppls2.GetPoint2dAt(0))
    return temppls2
  }

  /** */
  private static checkPoint(Block: PlaceBlock, x: number, y: number): boolean
  {
    if (x < 0 || y < 0)
      return false
    if (Block.Width < x || Block.Length < y)
      return false
    return true
  }

  /** 判断是否包含点位 */
  private getIntersectBox(nodeBox: NoteBox, graphys: Graphy[]): Graphy[]
  {
    let reArrPt: Graphy[] = []
    for (let grapy of graphys)
    {
      if (grapy.coverBox(nodeBox.Box))
        reArrPt.push(grapy)
    }
    return reArrPt
  }

  /** 获得边框点阵 (弧形拆封3段直线)   */
  private getBorderPoints(block: PlaceBlock): Point[]
  {
    let points: Point[] = []
    if (block.IsScrapBlock)
    {
      let ps = block.ScrapPoints
      for (let pt of ps)
      {
        points.push(new Point(pt.x, pt.y))
      }
      // 闭合
      points.push(new Point(points[0].X, points[0].Y))
    } else
    {
      let curves = BlockPlus.getOrgBorder(block, false)// BlockHelper.GetBorder_Org(block);
      let borderPoints = curves
      for (let i = 0; i < borderPoints.length; i++)
      {
        let p0 = borderPoints[i]
        // 如果是圆弧,拆开3段
        if (p0 instanceof Arc2d)
        {
          let arc = p0// new Arc2d(new Point2d(p0.PointX, p0.PointY), new Point2d(p1.PointX, p1.PointY), p0.Curve);
          let arcs = CADExt.SplitArc(arc, 4)
          for (let L of arcs)
          {
            points.push(new Point(L.StartPoint.m_X, L.StartPoint.m_Y))
          }
        }
        else // 直线
        {
          points.push(new Point(p0.StartPoint.m_X, p0.StartPoint.m_Y))
        }
      }

      // 闭合
      points.push(new Point(points[0].X, points[0].Y))
    }

    return points
  }
}

class Box
{
  public readonly minPt: CfPoint
  public readonly maxPt: CfPoint

  get W() { return this.maxPt.X - this.minPt.X }

  get L() { return this.maxPt.Y - this.minPt.Y }

  private constructor()
  {
    this.minPt = new CfPoint(Number.MAX_VALUE, Number.MAX_VALUE)
    this.maxPt = new CfPoint(Number.MIN_VALUE, Number.MIN_VALUE)
  }

  /** 通过点数组创建盒子 */
  public static createBox(minPt: CfPoint, maxPt: CfPoint): Box
  {
    let rect: Box = new Box()
    rect.minPt.X = minPt.X
    rect.minPt.Y = minPt.Y
    rect.maxPt.X = maxPt.X
    rect.maxPt.Y = maxPt.Y
    return rect
  }

  /** 通过点数组创建盒子 */
  public static createBoxByBoxs(minPt: Graphy[]): Box
  {
    let rect: Box = new Box()
    for (let pBox of minPt)
    {
      rect.AddBox(pBox.Box)
    }
    return rect
  }

  /** 通过点数组创建盒子 */
  public static createBoxByPoints(points: CfPoint[]): Box
  {
    let rect: Box = new Box()
    for (let pt of points)
    {
      rect.AddPoint(pt)
    }
    return rect
  }

  /** 通过点数组创建盒子 */
  public static createBoxByPoint(point: CfPoint, extentVal: number): Box
  {
    let rect: Box = new Box()
    rect.minPt.X = point.X - extentVal
    rect.minPt.Y = point.Y - extentVal
    rect.maxPt.X = point.X + extentVal
    rect.maxPt.Y = point.Y + extentVal
    return rect
  }

  public getDiagonalLength(): number
  {
    return this.minPt.DistensTo(this.maxPt)
  }

  /** 获取中点 */
  public getCenterPt(): CfPoint
  {
    return this.maxPt.Add(this.minPt).Multiply(0.5)
  }

  //* * 获取4个点 左下，右下，右上，左上  */
  public get4Pt(): CfPoint[]
  {
    let rePoints: CfPoint[] = []
    rePoints.push(new CfPoint(this.minPt.X, this.minPt.Y))
    rePoints.push(new CfPoint(this.maxPt.X, this.minPt.Y))
    rePoints.push(new CfPoint(this.maxPt.X, this.maxPt.Y))
    rePoints.push(new CfPoint(this.minPt.X, this.maxPt.Y))

    return rePoints
  }

  /** 扩大和延伸盒子 */
  private AddPoint(pt: CfPoint)
  {
    if (pt.X < this.minPt.X)
      this.minPt.X = pt.X
    if (pt.X > this.maxPt.X)
      this.maxPt.X = pt.X
    if (pt.Y < this.minPt.Y)
      this.minPt.Y = pt.Y
    if (pt.Y > this.maxPt.Y)
      this.maxPt.Y = pt.Y
  }

  /** 扩大和延伸盒子 */
  public AddBox(pBox: Box)
  {
    this.AddPoint(pBox.minPt)
    this.AddPoint(pBox.maxPt)
  }

  public extendBox(valw: number, vall: number)
  {
    this.minPt.X = this.minPt.X - valw
    this.maxPt.X = this.maxPt.X + valw
    this.minPt.Y = this.minPt.Y - vall
    this.maxPt.Y = this.maxPt.Y + vall
  }

  /** 判断是否包含点位 */
  public includesPt(x: number, y: number): boolean
  {
    if (x + 0.1 > this.minPt.X && x - 0.1 < this.maxPt.X)
    {
      if (y + 0.1 > this.minPt.Y && y - 0.1 < this.maxPt.Y)
      {
        return true
      }
    }
    return false
  }

  /** 判断是否包含点位 */
  public onLinePt(x: number, y: number): boolean
  {
    let miny = this.minPt.Y - 0.1
    let maxy = this.maxPt.Y + 0.1
    if (Math.abs(this.minPt.X - x) < 0.1 && (y > miny && y < maxy))
      return true
    if (Math.abs(this.maxPt.X - x) < 0.1 && (y > miny && y < maxy))
      return true

    let minx = this.minPt.X - 0.1
    let maxx = this.maxPt.X + 0.1
    if (Math.abs(this.minPt.Y - y) < 0.1 && (x > minx && x < maxx))
      return true
    if (Math.abs(this.maxPt.Y - y) < 0.1 && (x > minx && x < maxx))
      return true
  }

  /** 判断是否相交 */
  public intersect(box: Box): boolean
  {
    let tempMinPt = box.minPt
    let tempMaxPt = box.maxPt
    if (tempMinPt.X + 0.1 > this.maxPt.X)
      return false
    if (tempMinPt.Y + 0.1 > this.maxPt.Y)
      return false
    if (tempMaxPt.X + 0.1 < this.minPt.X)
      return false
    if (tempMaxPt.Y + 0.1 < this.minPt.Y)
      return false

    return true
  }
}
