import { FixIndex, equaln } from '../Util'
import type { Path } from './Path'

/**
 * 轮廓路径构造器
 * 传递一组简化后的点表过来，如果已经有同样的点表时，返回已经生产的Path，避免重复产生Path。
 * 使用相同的PATH有复用路径缓存。
 *
 * 每次进行优化时，必须清理构造器，保证Path生成是对本次优化唯一。
 */
class PathGenerator
{
  paths: Path[] = []
  pathAreaMap: { [key: string]: Path[] } = {}

  // 缓存命中次数
  cacheCount = 0

  /**
   * 如果存在同样的轮廓，则返回已经构造的轮廓，
   * 如果没有，则返回自身，并且注册它。
   * 如果id没有被注册,那么证明它无法放置在bin中
   */
  Allocate(path: Path): Path
  {
    let area = path.Area.toFixed(0)
    let paths = this.pathAreaMap[area]
    if (paths)
    {
      for (let ps of paths)
      {
        if (EqualPath(ps, path))
        {
          this.cacheCount++
          return ps
        }
      }
      paths.push(path)
    }
    else
      this.pathAreaMap[area] = [path]
    return path
  }

  RegisterId(path: Path)
  {
    if (path.Id === undefined)
    {
      path.Id = this.paths.length
      this.paths.push(path)
    }
  }

  Clear()
  {
    this.paths = []
    this.pathAreaMap = {}
    this.cacheCount = 0
  }
}

/**
 * 两路径相等,点表个数相等且每个点都相似
 */
function EqualPath(path1: Path, path2: Path): boolean
{
  if (path1.Points.length !== path2.Points.length)
    return false

  let p0 = path1.Points[0]
  let p2Index = path2.Points.findIndex((p) =>
  {
    return equaln(p.x, p0.x, 1e-3) && equaln(p.y, p0.y, 1e-3)
  })

  for (let i = 0; i < path1.Points.length; i++)
  {
    let p1 = path1.Points[i]
    let p2 = path2.Points[FixIndex(p2Index + i, path2.Points)]

    if (!equaln(p1.x, p2.x, 1e-4) || !equaln(p1.y, p2.y, 1e-4))
      return false
  }
  return true
}

export const PathGeneratorSingle = new PathGenerator()
